#!/usr/bin/env python

import sys, os, shutil, tempfile
import re
import random
import logging
import subprocess
from cStringIO import StringIO
import binascii
import Image
import xml.dom.minidom
import urllib
import zipfile, gzip, bz2
import optparse

VERSION = '0.1.bzr318'
VERSION_XSLT = '0.1.bzr320'

has_translify = False
try:
    import pytils
    from BeautifulSoup import BeautifulStoneSoup
    def translify(s):
        d = unicode(BeautifulStoneSoup(s, convertEntities = BeautifulStoneSoup.XML_ENTITIES))
        return pytils.translit.translify(d)
    has_translify = True
except ImportError, err: pass


registred_processors = {}
def register_processor(proc_class):
    registred_processors[proc_class.get_reg_name()] = proc_class


to_xml_uri = lambda x: x

def unix_to_xml_uri(x):
    x = urllib.pathname2url(x)
    return 'file://' + x

def win32_to_xml_uri(x):
    if len(x) > 2 and x[1] == ':':
        return 'file:///' + x[:2] + urllib.pathname2url(x[2:])
    else:
        return 'file://' + urllib.pathname2url(x)

if sys.platform == 'win32':
    to_xml_uri = win32_to_xml_uri
else:
    to_xml_uri = unix_to_xml_uri

class XSLTransformerException(Exception): pass
class XSLTLibxsltCommandException(XSLTransformerException): pass
class XSLTXalanCommandException(XSLTransformerException): pass
class XSLTSaxon6CommandException(XSLTransformerException): pass
class FOTransformerException(Exception): pass
class FOPCommandException(FOTransformerException): pass
class XEPCommandException(FOTransformerException): pass
class PdfToolException(Exception): pass
class PdftkCommandException(PdfToolException): pass
class PdftkCommandDumpDataException(PdftkCommandException): pass
class PdftkCommandUpdateInfoException(PdftkCommandException): pass
class PdfOptimizerException(Exception): pass
class GSCommandException(PdfOptimizerException): pass

def gen_tmp_name(prefix = '.fb2d'):
    return '%s%012X' % (prefix, random.randint(0, 0xFFFFFFFFFFFFL))

def str2bool(val):
    smap = {'true': True, 'false': False}
    try: return bool(int(val))
    except ValueError, err: pass
    return smap.get(val.lower(), None)

def import_module(mod_name, to_reload = False):
    ##todo: import module like mod.submod1.submod2.x.x
    module = None
    if sys.modules.has_key(mod_name):
        if to_reload:
            module = reload(sys.modules[mod_name])
            globals()[mod_name] = module
    else:
        module = __import__(mod_name, globals(), locals(), [])
        globals()[mod_name] = module
    return module

def findAll(elem, tagname):
    for node in elem.childNodes:
        if node.nodeType == node.ELEMENT_NODE and node.tagName == tagname:
            yield node

def getText(nodelist):
    rc = ""
    for node in nodelist:
        if node.nodeType == node.TEXT_NODE:
            rc = rc + node.data
    return rc

def setText(text, nodelist):
#set text to first text node in list!!!
    for node in nodelist:
        if node.nodeType == node.TEXT_NODE:
            node.data = text

def get_param(root_node, *node_names):
    if len(node_names) == 0:
        return getText(root_node.childNodes)
    name = node_names[0]
    for node in root_node.childNodes:
        if node.nodeType == node.ELEMENT_NODE and node.tagName == name:
            return get_param(node, *node_names[1:])

def set_param(root_node, text, *node_names):
    print root_node
    if len(node_names) == 0:
        if len(root_node.childNodes) == 0: root_node.appendChild(xml.dom.minidom.Text())
        return setText(text, root_node.childNodes)
    name = node_names[0]
    for node in root_node.childNodes:
        if node.nodeType == node.ELEMENT_NODE and node.tagName == name:
            return set_param(node, text, *node_names[1:])
    node = root_node.appendChild(xml.dom.minidom.Element(name))
    return set_param(node, text, *node_names[1:])

class Options(object):
    def __init__(self, opts_fname):
        self._soup = None
        self.load(opts_fname)

        self.image_transfrom = False        
        self.thumb_type = Image.ANTIALIAS
        self.cnv_tmp_dir = tempfile.gettempdir()

    def load(self, opts_fname):
        finp = None
        try:
            finp = file(opts_fname)
            self._soup = xml.dom.minidom.parse(finp)
        finally:
            if finp != None: finp.close()

    def save(self, opts_fname):
        ###add my exception, if _soup is None
        fout = None
        try:
            fout = file(opts_fname, 'wb')
            self._soup.writexml(fout)
        finally:
            if fout != None: fout.close()

    def option(self, *params):
        return get_param(self._soup.documentElement, *params).encode()
    def set_option(self, text, *params):
        return set_param(self._soup.documentElement, text, *params)

class XSLT_libxslt_exec(object):
    retcodes = {1: 'No argument',
                2: 'Too many parameters',
                3: 'Unknown option', 
                4: 'Failed to parse the stylesheet',
                5: 'Error in the stylesheet',
                6: 'Error in one of the documents',
                7: 'Unsupported xsl:output method',
                8: 'String parameter contains both quote and double-quotes',
                9: 'Internal processing error',
                10: 'Processing was stopped by a terminating message', }
    @staticmethod
    def get_reg_name(): return 'libxslt_bin'
    @staticmethod
    def use_type(): return 'exec'
    @staticmethod
    def get_document_element_type(): return 'exsl:document'
    def __init__(self, options):
        self._command = options.option('xslt_proc', 'libxslt_bin', 'bin')

    def transform(self, xslt_fname, fname_in, fname_out, params = tuple()):
        def yld_pars(command_name, pars, *other):
            yield command_name
            for p in pars:
                yield '--stringparam'
                yield p[0]
                yield p[1]
            for p in other: yield p
        args = list(yld_pars(self._command, params, '-output', fname_out, xslt_fname, fname_in))
        p = subprocess.Popen(args = args)
        retcode = p.wait()
        if retcode != 0:
            raise XSLTLibxsltCommandException('%s: %s' % (retcode, XSLT_libxslt_exec.retcodes.get(retcode, '')))
register_processor(XSLT_libxslt_exec)

class XSLT_libxslt_lib(object):
    @staticmethod
    def get_reg_name(): return 'libxslt_lib'
    @staticmethod
    def use_type(): return 'lib'
    @staticmethod
    def get_document_element_type(): return 'exsl:document'
    def __init__(self, options):
        import_module('libxml2')
        import_module('libxslt')
    def transform(self, xslt_fname, fname_in, fname_out, params = tuple()):
        def mk_params_dict(params):
            res = {}
            for par in params: res[par[0]] = "'%s'" % (par[1], )
            if len(res) == 0: res = None
            return res
        styledoc = libxml2.parseFile(xslt_fname)
        style = doc = result = None
        try:
            style = libxslt.parseStylesheetDoc(styledoc)
            doc = libxml2.parseFile(fname_in)
            params_dict = mk_params_dict(params)
            result = style.applyStylesheet(doc, params_dict)
            style.saveResultToFilename(fname_out, result, 0)
        finally:
            if style != None: style.freeStylesheet()
            if doc != None: doc.freeDoc()
            if result != None: result.freeDoc()
register_processor(XSLT_libxslt_lib)

class XSLT_4xslt_lib(object):
    @staticmethod
    def get_reg_name(): return '4xslt_lib'
    @staticmethod
    def use_type(): return 'lib'
    @staticmethod
    def get_document_element_type(): return 'exsl:document'
    def __init__(self, options):
        import_module('Ft.Xml.XPath')
        import_module('Ft.Xml.Xslt')
    def transform(self, xslt_fname, fname_in, fname_out, params = tuple()):
        def mk_params_dict(params):
            res = {}
            for par in params: res[par[0]] = par[1]
            if len(res) == 0: res = None
            return res
        params_dict = mk_params_dict(params)        
        fout = file(fname_out, 'wb')
        try:
            Ft.Xml.Xslt.Transform(fname_in, xslt_fname, params_dict, output=fout)
        finally:
            fout.close()
register_processor(XSLT_4xslt_lib)

class XSLT_xalan_exec(object):
    @staticmethod
    def get_reg_name(): return 'xalan_bin'
    @staticmethod
    def use_type(): return 'exec'
    @staticmethod
    def get_document_element_type(): return 'redirect:write'
    def __init__(self, options):
        self._command = options.option('xslt_proc', 'xalan_bin', 'bin')

    def transform(self, xslt_fname, fname_in, fname_out, params = tuple()):
        def yld_pars(command_name, pars, *other):
            yield command_name
            for p in pars:
                yield '-param'
                yield p[0]
                yield p[1]
            for p in other: yield p
        args = list(yld_pars(self._command, params, '-out', fname_out,
                             '-xsl', xslt_fname, '-in', fname_in))
        p = subprocess.Popen(args = args)
        retcode = p.wait()
        if retcode != 0:
            raise XSLTXalanCommandException(str(retcode))
register_processor(XSLT_xalan_exec)

class XSLT_saxon6_exec(object):
    @staticmethod
    def get_reg_name(): return 'saxon6_bin'
    @staticmethod
    def use_type(): return 'exec'
    @staticmethod
    def get_document_element_type(): return 'xsl:document'
    def __init__(self, options):
        self._command = options.option('xslt_proc', 'saxon6_bin', 'bin')

    def transform(self, xslt_fname, fname_in, fname_out, params = tuple()):
        def yld_pars(command_name, pars, *other):
            yield command_name
            for p in other: yield p
            for p in pars:
                yield '%s=%s' % p
        args = list(yld_pars(self._command, params, '-o', fname_out,
                             fname_in, xslt_fname, ))
        p = subprocess.Popen(args = args)
        retcode = p.wait()
        if retcode != 0:
            raise XSLTSaxon6CommandException(str(retcode))
register_processor(XSLT_saxon6_exec)

class FO_fop_exec(object):
    @staticmethod
    def get_reg_name(): return 'fop_bin'
    @staticmethod
    def use_type(): return 'exec'
    def __init__(self, options, command = None):
        self._command = command
        if command == None:
            self._command = options.option('fo_proc', 'fop_bin', 'bin')
        self._fop_config = os.path.abspath(options.option('fo_proc', 'fop_bin', 'config'))
        #self._dpi = options.output_dpi[0]
        self._dpi = None
    @staticmethod
    def docbook_params(): return (('fop1.extensions', '1'), )
    def transform(self, fo_fname, pdf_fname):
        args = [self._command, ]
        if self._fop_config: args += ['-c', self._fop_config]
        if self._dpi: args += ['-dpi', str(self._dpi)]
        args += ['-fo', fo_fname, '-pdf', pdf_fname]
        p = subprocess.Popen(args = args)
        retcode = p.wait()
        if retcode != 0:
            raise FOPCommandException(str(retcode))
register_processor(FO_fop_exec)

class FO_xep_exec(object):
    @staticmethod
    def get_reg_name(): return 'xep_bin'
    @staticmethod
    def use_type(): return 'exec'
    def __init__(self, options):
        self._command = options.option('fo_proc', 'xep_bin', 'bin')
    @staticmethod
    def docbook_params(): return (('xep.extensions', '1'), )
    def transform(self, fo_fname, pdf_fname):
        args = [self._command, ]        
        args += ['-fo', fo_fname, '-pdf', pdf_fname]
        p = subprocess.Popen(args = args)
        retcode = p.wait()
        if retcode != 0:
            raise XEPCommandException(str(retcode))
register_processor(FO_xep_exec)

class PdfTool_pdftk_exec(object):
    @staticmethod
    def get_reg_name(): return 'pdftk_bin'
    @staticmethod
    def use_type(): return 'exec'
    def __init__(self, options, command = None):
        self._command = command
        if command == None:
            self._command = options.option('pdf_proc', 'pdftk_bin', 'bin')
    def transform(self, soup_metadata, in_pdf_fname, out_pdf_fname,
                  in_dd_fname, out_dd_fname, translify_fun = lambda x: x):
        def dump_data(in_pdf_fname, in_dd_fname):
            args = [self._command, ]
            args += [in_pdf_fname, 'dump_data', 'output', in_dd_fname]
            p = subprocess.Popen(args = args)
            retcode = p.wait()
            if retcode != 0:
                raise PdftkCommandDumpDataException(str(retcode))
        def update_info(in_pdf_fname, out_pdf_fname, out_dd_fname):
            args = [self._command, ]
            args += [in_pdf_fname, 'update_info', out_dd_fname, 'output', out_pdf_fname]
            p = subprocess.Popen(args = args)
            retcode = p.wait()
            if retcode != 0:
                raise PdftkCommandUpdateInfoException(str(retcode))
        def write_new_metadata(soup_metadata, inp_dd_fname,
                               out_dd_fname, translify_fun):
            mts = []
            for node in soup_metadata.childNodes:
                if node.nodeType == node.ELEMENT_NODE:
                    mts.append((node.tagName, getText(node.childNodes)))
            finp = fout = None
            try:
                finp = file(in_dd_fname)
                fout = file(out_dd_fname, 'wb')
                for mt in mts:
                    s = 'InfoKey: %s\nInfoValue: %s\n' % tuple(map(lambda x: x.encode('ascii', 'xmlcharrefreplace'), mt))
                    fout.write(translify_fun(s))
                for line in finp:
                    fout.write(translify_fun(line))                
            finally:
                if finp != None: finp.close()
                if fout != None: fout.close()
        dump_data(in_pdf_fname, in_dd_fname)
        write_new_metadata(soup_metadata, in_dd_fname, out_dd_fname, translify_fun)
        update_info(in_pdf_fname, out_pdf_fname, out_dd_fname)
register_processor(PdfTool_pdftk_exec)

class PdfOptimizer_GS_exec(object):
    @staticmethod
    def get_reg_name(): return 'gspdfopt_bin'
    @staticmethod
    def use_type(): return 'exec'
    def __init__(self, options, command = None):
        self._command = command
        if command == None:
            self._command = options.option('pdf_optimizer', 'gspdfopt_bin', 'bin')
    def transform(self, inp_pdf_fname, out_pdf_fname):
        args = [self._command, inp_pdf_fname, out_pdf_fname]        
        p = subprocess.Popen(args = args)
        retcode = p.wait()
        if retcode != 0:
            raise GSCommandException(str(retcode))
register_processor(PdfOptimizer_GS_exec)

class Transformer(object):
    def __init__(self, options, xslt_transformer, fo_transformer,
                 pdf_tool, pdf_optimizer,
                 xslt_params, xslt_titlepages,
                 update_metadata = True, linearize_pdf = True, to_translify = False):
        self.options = options
        self.xslt_transformer = xslt_transformer
        self.fo_transformer = fo_transformer
        self.pdf_tool = pdf_tool
        self.pdf_optimizer = pdf_optimizer
        self.xslt_params = xslt_params
        self.xslt_titlepages = xslt_titlepages
        self.update_metadata = update_metadata
        self.linearize_pdf = linearize_pdf
        self.to_translify = to_translify
        self._tmp_fnames = set()
    def _add_tmp_fname(self, fname):
        self._tmp_fnames.add(os.path.abspath(fname))
    def _trasform_image(self, ascii_finp, bin_fout):        
        im = Image.open(StringIO(binascii.a2b_base64(ascii_finp.read())))
        max_size = self.options.max_image_size
        if max_size[0] < im.size[0] or max_size[1] < im.size[1]:
            t_ratio = min(max_size[0]/float(im.size[0]), max_size[1]/float(im.size[1]))
            im.thumbnail(map(lambda x: x*t_ratio, im.size), self.options.thumb_type)
        im.convert("L").save(bin_fout, dpi = self.options.output_dpi)
    def _wrk_transform_binaries(self, conv_info_soup):        
        self.options.image_transform = str2bool(get_param(conv_info_soup.documentElement, 'page', 'images_mode', 'resize'))
        if self.options.image_transform:
            self.options.image_mode = get_param(conv_info_soup.documentElement, 'page', 'images_mode', 'mode')
            self.options.output_dpi = (int(get_param(conv_info_soup.documentElement, 'page', 'dpi', 'width')),
                        int(get_param(conv_info_soup.documentElement, 'page', 'dpi', 'height')))
            page_size = (get_param(conv_info_soup.documentElement, 'page', 'size', 'width'),
                         get_param(conv_info_soup.documentElement, 'page', 'size', 'height'))
            self.options.page_size = tuple(map(lambda x: float(x[:-2]), page_size))
            self.options.page_image_margin = (int(get_param(conv_info_soup.documentElement, 'page', 'max_image_margin', 'width')),
                                 int(get_param(conv_info_soup.documentElement, 'page', 'max_image_margin', 'height')))
            self.options.max_image_size = ((((self.options.page_size[0] - self.options.page_image_margin[0]*2)/25.4)*self.options.output_dpi[0],
                ((self.options.page_size[1] - self.options.page_image_margin[1]*2)/25.4)*self.options.output_dpi[1]))            
        for binary_node in findAll(conv_info_soup.documentElement.getElementsByTagName('binaries')[0], 'binary'):
            href_binary = binary_node.attributes['href_binary'].value
            href_ascii = binary_node.attributes['href_ascii'].value
            content_type = binary_node.attributes['content-type'].value
            bin_fout = file(href_binary, 'wb')
            self._add_tmp_fname(href_binary)
            ascii_finp = file(href_ascii)
            self._add_tmp_fname(href_ascii)
            try:
                if self.options.image_transform and content_type.split('/')[0].lower() == 'image':
                    logging.info('transforming image: %s' % href_binary)
                    self._trasform_image(ascii_finp, bin_fout)
                else:
                    bin_fout.write(binascii.a2b_base64(ascii_finp.read()))
            finally:
                bin_fout.close()
                ascii_finp.close()
    def unpack_try(self, inp_fb2_fname):
        def unp_gzip(fnp, fu):
            gf = gzip.GzipFile(fnp)
            try:
                fu.write(gf.read())
            finally:
                gf.close()
        def unp_bzip2(fnp, fu):
            gf = bz2.BZ2File(fnp)
            try:
                fu.write(gf.read())
            finally:
                gf.close()
        def unp_zip(fnp, fu):
            try:
                ar = zipfile.ZipFile(fnp)
                namelist = ar.namelist()
                if len(namelist) == 1: fn_u = namelist[0]
                else: fn_u = os.path.splitext(os.path.basename(fnp))[0]
                fu.write(ar.read(fn_u))
            finally:
                ar.close()
        ext_map = {'.gz': unp_gzip,
                  '.zip': unp_zip,
                  '.bz2': unp_bzip2
                  }
        name, ext = os.path.splitext(os.path.basename(inp_fb2_fname))
        unp_fun = ext_map.get(ext.lower(), None)
        if unp_fun == None:
            logging.info('file seems to be non-packed (has no known packer extesnsion)')
            return inp_fb2_fname
        logging.info('file seems to be packed (has known packer extesnsion)')
        try:
            unp_inp_fb2_fname = os.path.join(self.trns_dir, name)        
            fu = file(unp_inp_fb2_fname, 'wb')
            self._add_tmp_fname(unp_inp_fb2_fname)
            logging.info('unpacking file')
            unp_fun(inp_fb2_fname, fu)
        finally:
            fu.close()
        return unp_inp_fb2_fname
    def transform(self, fb2_fname, pdf_fname):
        self.cur_dir = os.getcwd()
        self.trns_dir = os.path.abspath(os.path.join(self.options.cnv_tmp_dir, gen_tmp_name()))
        try:
            os.mkdir(self.trns_dir)
        except OSError, err:
            if err.errno != 17: raise err

        inp_fb2_fname = os.path.abspath(fb2_fname)
        inp_fb2_fname = self.unpack_try(inp_fb2_fname)
        out_pdf_fname = os.path.abspath(pdf_fname)
        self.interim_name = os.path.split(inp_fb2_fname)[1]        
        conv_info_fname = self.options.option('conv_info_fname')
        docbook_fname = self._transform_fb2_docbook(inp_fb2_fname, conv_info_fname)
        abs_conv_info_fname = os.path.abspath(os.path.join(self.trns_dir, conv_info_fname))
        self._add_tmp_fname(abs_conv_info_fname)
        conv_info_soup = xml.dom.minidom.parse(abs_conv_info_fname)
        self._transform_binaries(conv_info_soup)
        titlepages_fname = self._transform_titlepages()
        abs_fo_fname = self._transform_docbook_fo(docbook_fname, titlepages_fname)
        fo_pdf_fname = self._transform_fo(abs_fo_fname)
        wf_pdf_fname = fo_pdf_fname

        if self.update_metadata:            
            upd_desc_pdf_fname = self._make_pdf_description(
                                    conv_info_soup.documentElement.getElementsByTagName('metadata')[0],
                                    wf_pdf_fname)
            self._add_tmp_fname(wf_pdf_fname)
            wf_pdf_fname = upd_desc_pdf_fname
        if self.linearize_pdf:            
            optimzed_pdf_fname = self._pdf_optimize(wf_pdf_fname)
            self._add_tmp_fname(wf_pdf_fname)
            wf_pdf_fname = optimzed_pdf_fname
        logging.info('moving resulting pdf to final location')
        shutil.move(wf_pdf_fname, out_pdf_fname)
    def _transform_fo(self, abs_fo_fname):
        try:
            logging.info('transforming XSL-FO to pdf')
            os.chdir(self.trns_dir)
            fo_pdf_fname = os.path.abspath(self.interim_name + '.pdf')
            self.fo_transformer.transform(abs_fo_fname, fo_pdf_fname)
        finally:
            os.chdir(self.cur_dir)
        return fo_pdf_fname
    def _pdf_optimize(self, inp_pdf_fname):
        try:
            logging.info('linearizing pdf')
            out_pdf_fname = os.path.abspath(os.path.join(self.trns_dir, self.interim_name + '.linearized.pdf'))
            self.pdf_optimizer.transform(inp_pdf_fname, out_pdf_fname)
        finally:
            pass
        return out_pdf_fname    
    def _make_pdf_description(self, soup_metadata, in_pdf_fname):        
        try:
            logging.info('updating pdf description')
            in_dd_fname = os.path.abspath(os.path.join(self.trns_dir, self.interim_name + '.inp.metadata'))
            self._add_tmp_fname(in_dd_fname)
            out_dd_fname = os.path.abspath(os.path.join(self.trns_dir, self.interim_name + '.out.metadata'))
            self._add_tmp_fname(out_dd_fname)
            out_pdf_fname = os.path.abspath(os.path.join(self.trns_dir, self.interim_name + '.upd.desc.pdf'))
            translify_fun = lambda x: x
            if self.to_translify: translify_fun = translify
            self.pdf_tool.transform(soup_metadata, in_pdf_fname,
                                    out_pdf_fname, in_dd_fname, out_dd_fname,
                                    translify_fun)
        finally:
            pass
        return out_pdf_fname
    def _transform_binaries(self, conv_info_soup):
        logging.info('transforming binaries')
        try:
            os.chdir(self.trns_dir)
            self._wrk_transform_binaries(conv_info_soup)
        finally:
            os.chdir(self.cur_dir)
    def _transform_titlepages(self):
        logging.info('preparing xsl for titlepages')
        titlepages_fname_inp = os.path.abspath(self.xslt_titlepages)
        titlepages_fname = os.path.abspath(os.path.join(self.trns_dir,
                                                           os.path.split(titlepages_fname_inp)[1] + '.xsl'))
        xslt_trans = os.path.abspath(os.path.join(self.options.option('docbook', 'root'),
                                                  self.options.option('docbook', 'titlepages_trans')))
        self.xslt_transformer.transform(xslt_trans, titlepages_fname_inp, titlepages_fname)
        self._add_tmp_fname(titlepages_fname)
        return titlepages_fname
    def _transform_docbook_fo(self, docbook_fname, titlepages_fname):
        def cr_trns_f(fname):
            content = """<?xml version='1.0' encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform" version="1.1">
            <xsl:import href="%s"/>
            <xsl:import href="%s"/>
            <xsl:import href="%s"/>
</xsl:stylesheet>"""
            fout = file(fname, 'wb')
            try:
                docbook_xslt = os.path.abspath(os.path.join(self.options.option('docbook', 'root'),
                                                            self.options.option('docbook', 'fo_trans')))
                user_xslt = os.path.abspath(self.xslt_params)
                fout.write(content % (to_xml_uri(docbook_xslt), to_xml_uri(titlepages_fname), to_xml_uri(user_xslt)))
            finally:
                fout.close()
        logging.info('transforming Docbook to XSL-FO')
        trns_xslt_fname = os.path.abspath(os.path.join(self.trns_dir, 'trns_docbook_fo.xsl'))
        cr_trns_f(trns_xslt_fname)
        self._add_tmp_fname(trns_xslt_fname)
        abs_fo_fname = os.path.abspath(os.path.join(self.trns_dir, self.interim_name + '.fo'))
        self.xslt_transformer.transform(trns_xslt_fname,
                                        docbook_fname,
                                        abs_fo_fname,
                                        self.fo_transformer.docbook_params())
        self._add_tmp_fname(abs_fo_fname)
        return abs_fo_fname
    def _transform_fb2_docbook(self, inp_fb2_fname, conv_info_idx_fname):
        def cr_trns_f(fname):
            content = """<?xml version='1.0' encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform" version="1.1">
            <xsl:import href="%s"/>
            <xsl:include href="%s"/>
</xsl:stylesheet>"""
            fout = file(fname, 'wb')
            try:
                user_xslt = os.path.abspath(self.xslt_params)
                script_xslt = os.path.abspath(os.path.join(self.options.option('fb2', 'root'),
                    self.options.option('fb2', 'docbook_trans')))
                fout.write(content % (to_xml_uri(user_xslt), to_xml_uri(script_xslt)))
            finally:
                fout.close()
        params = (('conv_info_idx', conv_info_idx_fname),
                  ('document-element', self.xslt_transformer.get_document_element_type()))
        try:
            logging.info('transforming FB2 to Docbook')
            trns_xslt_fname = os.path.join(self.trns_dir, 'trns_fb2_docbook.xsl')
            cr_trns_f(trns_xslt_fname)
            self._add_tmp_fname(trns_xslt_fname)
            os.chdir(self.trns_dir)
            docbook_fname = os.path.abspath(self.interim_name + '.docbook')
            self.xslt_transformer.transform(trns_xslt_fname, inp_fb2_fname,
                                            docbook_fname, params)
            self._add_tmp_fname(docbook_fname)
        finally:
            os.chdir(self.cur_dir)
        return docbook_fname
    def rm_tmp_files(self):
        for fn in self._tmp_fnames:
            try:
                os.unlink(fn)
            except Exception, err:
                logging.error('Error %s while unlinking file: %s', err, fn)
        try:
            os.rmdir(self.trns_dir)
        except Exception, err:
            logging.error('Error %s while removing directory: %s', err, self.trns_dir)

def setup_logger():
    rootLogger = logging.getLogger('')
    rootLogger.setLevel(logging.INFO)
    formatter = logging.Formatter('%(asctime)s %(levelname)s: %(message)s')
    lg_hnd = logging.StreamHandler(sys.stdout)
    lg_hnd.setFormatter(formatter)
    rootLogger.addHandler(lg_hnd)

def get_pdf_fname(fb2_fname):
    r_p = r"(\.fb2)?(\.(?:zip|gz|bz2))?$"
    return '%s.pdf' % (fb2_fname[:re.search(r_p, fb2_fname).start()], )

if __name__ == '__main__':
    setup_logger()
    prog_cwd = os.path.dirname(sys.argv[0])
    cur_cwd = os.getcwd()    
    cl_parser = optparse.OptionParser(usage = 'usage: %prog --config config_file [options] fb2_file[.(zip|gz|bz2)] pdf_file',
                             version = '%s version %s; fb2docbook.xsl version %s' % ('%prog', VERSION, VERSION_XSLT))    
    cl_parser.add_option('-c', '--config', dest = 'configname',
                  help = 'read config from FILE', metavar = 'FILE')
    cl_parser.add_option('--wh', action='store_true', dest='win32hack',
                         help = 'turn on win32hack -- non closed console window')
    cl_parser.add_option('--no-remove_tmp', action='store_false', default=True, dest='remove_tmp',
                         help = 'do not remove working temporal files')
    cl_parser.add_option('--no-update_metadata', action='store_false', default=True, dest='update_metadata',
                         help = 'do not update metadata')
    cl_parser.add_option('--no-linearize_pdf', action='store_false', default=True, dest='linearize_pdf',
                         help = 'do not linearize pdf')
    cl_parser.add_option('--xslt_proc', action='store', type='string', dest='xslt_proc',
                         help = 'set xslt_proc to use')
    cl_parser.add_option('--fo_proc', action='store', type='string', dest='fo_proc',
                         help = 'set fo_proc to use')
    cl_parser.add_option('--pdf_proc', action='store', type='string', dest='pdf_proc',
                         help = 'set pdf_proc to use')
    cl_parser.add_option('--pdf_optimizer', action='store', type='string', dest='pdf_optimizer',
                         help = 'set pdf_optimizer to use')
    cl_parser.add_option('--xslt_params', action='store', type='string', dest='xslt_params',
                         help = 'set xslt_params file to use', metavar='FILE')
    cl_parser.add_option('--xslt_titlepages', action='store', type='string', dest='xslt_titlepages',
                         help = 'set xslt_titlepages file to use', metavar='FILE')
    if has_translify:
        cl_parser.add_option('--translify', action='store_true', default=False, dest='to_translify',
                         help = 'translify pdf description')
    (cl_options, cl_args) = cl_parser.parse_args()

    if cl_options.win32hack:
        import atexit, msvcrt
        def getch():
            print 'Press any key...'
            msvcrt.getch()
        atexit.register(getch)

    logging.info('initializing')
    if not cl_options.configname:
        cl_options.configname = os.path.join(prog_cwd, 'config/btconfig.xml')
    if len(cl_args) > 2 or len(cl_args) == 0:
        cl_parser.error('you should provide at least input fb2_file')
    if len(cl_args) == 1:
        fb2_fname = cl_args[0]
        pdf_fname = get_pdf_fname(fb2_fname)
    else:
        fb2_fname, pdf_fname = cl_args

    options = Options(cl_options.configname)

    if not cl_options.xslt_proc:
        xslt_proc_name = options.option('xslt_proc', 'default')
    else:
        xslt_proc_name = cl_options.xslt_proc
    if not cl_options.fo_proc:
        fo_proc_name = options.option('fo_proc', 'default')
    else:
        fo_proc_name = cl_options.fo_proc
    if not cl_options.pdf_proc:
        pdf_proc_name = options.option('pdf_proc', 'default')
    else:
        pdf_proc_name = cl_options.pdf_proc
    if not cl_options.pdf_optimizer:
        pdf_optimizer_name = options.option('pdf_optimizer', 'default')
    else:
        pdf_optimizer_name = cl_options.pdf_optimizer

    xslt_proc = registred_processors[xslt_proc_name](options)    
    fo_proc = registred_processors[fo_proc_name](options)    
    pdf_proc = registred_processors[pdf_proc_name](options)
    pdf_optimizer = registred_processors[pdf_optimizer_name](options)

    logging.info('using xslt_proc: %s; fo_proc: %s; pdf_proc: %s, pdf_optimizer: %s',
                 xslt_proc_name, fo_proc_name, pdf_proc_name, pdf_optimizer_name)
        

    if not cl_options.xslt_params:
        xslt_params = options.option('xslt_params')
    else:
        xslt_params = cl_options.xslt_params

    if not cl_options.xslt_titlepages:
        xslt_titlepages = options.option('xslt_titlepages')
    else:
        xslt_titlepages = cl_options.xslt_titlepages

    trns = Transformer(options, xslt_proc, fo_proc,
                       pdf_proc, pdf_optimizer,
                       xslt_params, xslt_titlepages,
                       update_metadata = cl_options.update_metadata,
                       linearize_pdf = cl_options.linearize_pdf,
                       to_translify = has_translify and cl_options.to_translify)
    try:
        trns.transform(fb2_fname, pdf_fname)
    finally:
        if cl_options.remove_tmp:
            logging.info('removing temprorary files')
            trns.rm_tmp_files()
    logging.info('finish')

